/*
 * File: emotif-search.c
 * Date:  08-23-99
 * Programmers:  Thomas Wu (thomas.wu@Stanford.Edu)
 *               Craig Nevill-Manning (nevill@cs.rutgers.edu)
 * Comment Writer:  Jimmy Y. Huang (yhuang@leland.stanford.edu)
 * ------------------------------------------------------------
 * From the man page:
 * 
 * The  program  eMOTIF-SEARCH  is  a series of programs that
 * characterize a sequence by comparing it with a precompiled
 * database  of discrete motifs.  These discrete motifs, gen-
 * erated by the program eMOTIF-MAKER , from a set  of  input
 * files   specified  during  installation  in  the  variable
 * RAW_BLOCKS_INPUT    (blocks.dat)    and    the    variable
 * RAW_PRINTS_INPUT (prints.dat).  The output is ranked by p-
 * value.
 * ------------------------------------------------------------ 
 * Usage:  emotif-search [-s <sig>] [<sequencefile>]
 * ------------------------------------------------------------
 * Copyright 1999 The Board of Trustees of 
 * The Leland Stanford Junior University. All Rights Reserved.
 */


#include <ctype.h>
#include <math.h>
#include "sequence.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


/*
 * ----------
 * Function Prototypes
 * ----------
 */

static void InitializeMap (int *map);
static int *MapSequence (char *sequence_char, int length, int *map);
static void CopyHit (char *segment, char *sequence_char, int start,
		     int width, int length);


extern char *id[];
extern char *motif[];
extern char *sensitivity[];
extern int motif_index[];
extern unsigned long int ints[];
extern int total;
extern double specificity[];


/*
 * ----------
 * InitializeMap
 * ==========
 * The function InitializeMap initializes the int array the pointer
 * map of type int points to.  The int pointer map is a parameter given
 * to the function InitializeMap.  This int array must contain at
 * least 129 elements.  Each int element is essentially a bit vector
 * whose pattern is supposed to represent one of the 26 letters in the
 * English alphabet.
 * ----------
 */

static void InitializeMap (int *map)
{
  int i;
  char c;

  for (i = 0; i < 129; ++i) {

    /* 
     * ----------
     * The element map[i] is set to 0, so all the bits associated with
     * map[i] will be 0.  If map[i] is set to -1, everything will match,
     * because the two's complement representation of -1 is, for an 8-bit
     * number, 11111111.  Two's complement number representation is
     * explained below by Erik Haugen, a CS107 TA for the autumn of 1998.
     * 
     * Lets assume we have 4 bit words:
     *  3 is 0011
     *  2 is 0010
     *  1 is 0001
     *  0 is 0000
     * -1 is 1111
     * -2 is 1110
     * -3 is 1101
     * -4 is 1100
     *
     * Notice how it appears that we are just subtracting 1 from the
     * number each time.  The idea is that the machinery in the CPU
     * for adding can just treat negative numbers the same as it does
     * positive numbers.
     *
     * Adding 1111 + 0001 gives you 0000.  You lose the carry, because
     * we only have 4 bit words.  So it works great for 2's complement.
     *
     * A quick way to flip the sign of a 2's complement number is to
     * subtract 1 and then flip each bit.
     *
     *  1: 0001 => 0000 => 1111   -1
     *  2: 0010 => 0001 => 1110   -2
     *  5: 0101 => 0100 => 1011   -5
     * -3: 1101 => 1100 => 0011    3
     *
     * Now, if you tried to subtract 1 from 1000 (-8), you'd go out of
     * bounds, and get 0111, which is 7.  This kind of thing happens in
     * C if you keep subtracting 1.  (Similarly if you add 7 + 1, you
     * get -8.)
     * ----------
     */

    map[i] = 0;
  }  /* for (i = 0; i < 129; ++i) */


  /*
   * ----------
   * The following for loop will turn on bit 0 for the int variable
   * map['A'] (The ASCII character 'A' has a corresponding integer value
   * of 65.), bit 1 for the int variable map['B'] (The ASCII
   * character 'B' has a corresponding integer value of 66.), bit 2 for
   * the int variable map['C'] (The ASCII character 'C' has a
   * corresponding integer value of 67.), all the way to bit 25 for the
   * int variable map['Z'] (The ASCII character 'a' has a
   * corresponding integer value of 90.).
   * ----------
   */

  for (c = 'A', i = 0; c <= 'Z'; ++c, ++i) {
    map[c] = 1 << i;
  }  /* for (c = 'A', i = 0; c <= 'Z'; ++c, ++i) */


  /*
   * ----------
   * The following for loop will turn on bit 0 for the int variable
   * map['a'] (The ASCII character 'a' has a corresponding integer value
   * of 97.), bit 1 for the int variable map['b'] (The ASCII
   * character 'b' has a corresponding integer value of 98.), bit 2 for
   * the int variable map['c'] (The ASCII character 'c' has a 
   * corresponding integer value of 98.), all the way to bit 25 for the
   * integer var map['z'] (The ASCII character 'z' has a
   * corresponding integer value of 122.).
   * ----------
   */

  for (c = 'a', i = 0; c <= 'z'; ++c, ++i) {
    map[c] = 1 << i;
  }  /* for (c = 'a', i = 0; c <= 'z'; ++c, ++i) */
}


/*
 * ----------
 * MapSequence
 * ==========
 * The function MapSequence will convert the string of amino acid letters
 * in the dynamic array the char pointer sequence_char points to into a
 * string of bits.  
 * ----------
 */

static int *MapSequence (char *sequence_char, int length, int *map)
{
  int *sequence_int, i;


  /*
   * ----------
   * The function calloc, exported by stdlib.h, will return a pointer to
   * an array of length objects, each object is the size of an int.  This
   * function returns NULL, if the request cannot be satisfied.  The
   * function calloc will automatically initialize every object in that
   * array to zero.
   * ----------
   */

  sequence_int = (int *) calloc (length, sizeof (int));


  /*
   * ----------
   * The following for loop will initialize every element in the dynamic
   * array the int pointer sequence_int points to with the appropriate
   * bit patterns stored in the corresponding element in the array the
   * int pointer map points to.  The following example should illustrate
   * what Tom is doing clearly:
   *
   * > THTM_HUMAN\n
   * MASPQLCRAL VSAQWVAEAL RAPRAGQPLQ LLDASWYLPK LGRDATQFEE RHIPGAAFFD\n
   * IDQCSDRTSP YDHMLPGAEH FAEYAGRLGV GAATHVVIYD ASDQGLYSAP RVWWMFRAFG\n
   * HHAVSLLDGG LRHWLRQNLP LSSGKSQPAP AEFRAQLDPA FIKTYEDIKE NLESRRFQVV\n
   * DSRATGRFRG TEPEPRDGIE PGHIPGTVNI PFTDFLSQEG LEKSPEEIRH LFQEKKVDLS\n
   * KPLVATCGSG VTACHVALGA YLCGKPDVPI YDGSWVEWYM RARPEDVISE GRGKTH\n
   * 
   * The sequence is 296 amino acids long, so the variable length will
   * be set to 295.  The int pointer sequence_int points to a dynamic
   * int array of size length.  Every element in that dynamic int array
   * is essentially a bit vector with a corresponding position in the
   * amino acid sequence.  The char pointer sequence_char points to a
   * dynamic char array of size length, and each char element has
   * already been set to ASCII value of the corresponding position
   * in the input amino acid sequence.  The variable sequence_char[0],
   * for example, would be 'M', or 77; the variable sequence_char[1]
   * would be 'A', or 65; the variable sequence_char[2] would be 'S',
   * or 86, and etc.  The variable sequence_char[i], in turn, functions
   * as the index variable for the static array the int pointer map
   * points to.  At this point of the program elements 65 through 90 and
   * elements 97 through 122 in the static int array the pointer map
   * points to should have already been initialized with the appropriate
   * bit patterns representing the letters of the English alphabet in
   * uppercase and lowercase, respectively.  The bit pattern for the
   * amino acid alanine, 'A' or 'a', for example, is simply the bit pattern
   * for the integer 0; the bit pattern for the amino acid tyrosine, 'Y'
   * 'y', is simply the bit pattern for the integer 24.  For the current
   * example, the variable sequence_int[0], which corresponds to the
   * first methionine residue in the amino acid sequence, is ultimately
   * set to the bit pattern for the integer 12 ('M' is the 13th letter
   * of the English alphabet, but the counting started at 0.).
   * ----------
   */

  for (i = 0; i < length; ++i) {
    sequence_int[i] = map[sequence_char[i]];
  }  /* for (i = 0; i < length; ++i) */

  return (sequence_int);
}


/*
 * ----------
 * CopyHit
 * ==========
 * The function CopyHit is private to the file emotif-search.c.
 * This function requires five parameters.  It is basically responsible
 * for extracting the segment of input sequence that matches the current
 * eMOTIF under investigation in the main function and the fifteen
 * residues that precede and follow that segment.  The first parameter
 * is the char pointer segment that points to a static char array on the
 * main stack; the output of the CopyHit function is ultimately stored
 * in this static array.  The second parameter is the char pointer
 * sequence_char that points to the dynamic array containing the input
 * sequence.  The third parameter is the int variable start, which
 * spans from 0 to (length - 1).  The int variable start contains the
 * index number that corresponds to a char element in the static array
 * the char pointer sequence_char points to.  This char element that
 * the int variable start refers to contains the first amino acid letter
 * in the input sequence, whose every amino acid letter is stored in the
 * static array the char pointer sequence_char points to, that matches
 * the first position in the current eMOTIF.  The int variable width
 * lets the private function CopyHit know just how long the current
 * eMOTIF is, so it can calculate how many amino acid letters to extract
 * from the static array the char pointer sequence_char points to before
 * stopping.  The int variable length stores the length of the input
 * sequence.  The extracted amino acid letters are put in the static
 * array the char pointer segment points to.
 * ----------
 */

static void CopyHit (char *segment, char *sequence_char, int start,
		      int width, int length)
{

  /*
   * ----------
   * The int variable l is an index variable for the static array the
   * char pointer sequence_char points to.  The int variable p is an
   * index variable for the static array the char pointer segment
   * points to.
   * ----------
   */

  int l, p;


  l = start;
  p = 0;

  segment[p++] = sequence_char[l++];


  /*
   * ----------
   * The while loop below puts the actual amino acid letters that match
   * the current eMOTIF in the input sequence into the static array the
   * char pointer segment points to.
   *
   * Tom originally wrote:
   *
   * for ( ; l < start + width; ++l) {
   *   segment[p++] = sequence_char[l];
   * }
   *
   * A while loop, however, makes the code look cleaner.
   * ----------
   */

  while (l < start + width) {
    segment[p++] = sequence_char[l++];
  }  /* while (l < start + width) */

  segment[p] = '\0';
}


int main (int argc, char *argv[])
{
  FILE *fp;
  char *seqfile = NULL, c, *sequence_char, *name, segment[1000];
  int *sequence_int, motifn, length, width, i, j, k;
  int map[129];


  /*
   * ----------
   * The constant DEFAULT_SEARCH_THRESHOLD is set to 7 in the file
   * site.defs.  If the following statement is executed:
   *
   *      cc -DDEFAULT_SEARCH_THRESHOLD=7 -c foo.c
   *
   * The C macro preprocessor will substitute the value 7 wherever it sees
   * DEFAULT_SEARCH_THRESHOLD, just as the statement,
   * #define DEFAULT_SEARCH_THRESHOLD 7, would.  The following snippet is
   * taken from the cc man page:
   *
   * -Dname=def
   * -Dname
   *      Define the name to the C macro preprocessor, as if by `#define'.
   *      If no definition is given, the name is defined as "1".
   *
   * ----------
   */

  int opt, cutoff_specificity = 7;


  /*
   * ----------
   * The extern variable optind of type int is exported by the unistd.h
   * library.  The unistd.h function getopt places in optind the argv
   * index of the next argument to be processed.  The variable optind is
   * initialized to 1 before the first call to the function getopt.  When
   * all options have been processed (i.e., up to the first non-option
   * argument), getopt returns -1.
   * ----------
   */

  extern int optind;
  

  /*
   * ----------
   * The function getopt is a function the file unistd.h exports.  The
   * prototype of the getopt function is shown below:
   *
   *      int getopt (int argc, char * const argv[], const char *optstring);
   *
   * The getopt function returns the next option letter in the command-line
   * string the pointer argv points to that matches a letter in optstring.
   * The variable optstring is, again, a pointer to a constant character
   * object.  While the value of optstring may change, the value of whatever 
   * optstring points to may not.  The pointer opstring must point to a string
   * that contains the option letters the command using getopt will recognize.
   * If a letter is followed by a colon, the option is expected to have an
   * argument, which may be separated from it by a white space.  The extern
   * variable optarg of type char* points to the argument expected by an
   * option.  The function getopt is essentially responsible for directing the
   * exported pointer optarg of type char to the user-supplied argument.  The
   * function getopt returns -1 when all command line options have been parsed.
   * ----------
   */

  while ((opt = getopt (argc, argv, "s:")) != -1) {
    switch (opt) {

      /*
       * ----------
       * The function atoi is a function exported by the file stdlib.h.  The
       * atoi function takes a pointer to a constant character object and
       * returns the integer value represented by that constant character
       * object.  The following line of code is an example of the function
       * atoi's usage:
       *
       *      int integerValue = atoi ("8");
       *
       * The value of integerValue will be 8.  If the user has not set a
       * cut-off specificity value, the default value of "8" set earlier
       * will remain.
       * ----------
       */

    case 's': cutoff_specificity = atoi (optarg);
      break;
    }  /* switch (opt) */
  }  /* while ((opt = getopt (argc, argv, "s:")) != -1) */


  /*
   * ----------
   * The variable argc of type int is the argument count.  The
   * statement
   *
   *      emotif-search -s 10 pdb.fasta
   *
   * is broken up into 4 parts:  argv[0] = "emotif-search",
   * argv[1] = "-s", argv[2] = "10", and argv[3] = "pdb.fasta".
   * The value of argc, therefore, is 4.  Since the first non-option
   * argument encountered is pdb.fasta, the extern variable optind is 3.
   * The program  emotif-search then knows to expect a sequence file.
   * If no sequence filename is given, the extern variable optind is
   * guaranteed to be at least the value of the variable argc.  The
   * emotif-search program then knows to expect the sequence from the
   * standard input (stdin) directly.
   * ----------
   */

  if (optind >= argc) {
    fp = stdin;
  } else /* if (optind >= argc) */ {
    seqfile = argv[optind];
 
    if (!(fp = fopen (seqfile,"r"))) {
      fprintf (stderr, "Can't open sequence file %s\n", seqfile);

      exit (1);
    }  /* if (!(fp = fopen (seqfile,"r"))) */
  }  /* if (optind >= argc) */


  /*
   * ----------
   * The variable map is a pointer to a static int array of size 129.  The
   * function InitializeMap will initialize the static array the pointer
   * map points to appropriately.
   * ----------
   */

  InitializeMap (map);


  /*
   * ----------
   * The program will remain within the while loop below as long as
   * two conditions are satisfied:
   * 
   *      1.  The EOF character has not been reached;
   *      2.  The current character is not a space character.
   *
   * ----------
   */

  while (!feof (fp) && !isspace (c = fgetc (fp))) {


    /* 
     * ----------
     * The current character must be a non-space character, and it must
     * be placed back into the file stream.
     * ----------
     */

    ungetc (c,fp);


    /*
     * ----------
     * The function Sequence_read, exported by the file sequence.h, takes
     * three parameters.  The first parameter is the FILE pointer fp; the
     * second parameter is the char pointer name; and the third parameter
     * is the address of the int variable length.  The function
     * Sequence-read returns the char pointer sequence_char.  If the
     * char pointer sequence_char is not NULL, the char pointer name will
     * be pointing to a dynamic char array, and the int variable length
     * will be initialized with the length of the input sequence.
     * ----------
     */

    if (sequence_char = Sequence_read (fp, &name, &length)) {


      /*
       * ----------
       * If the char pointer sequence_char is not NULL, the char pointer
       * name will be pointing to a dynamic char array, and the int
       * variable length will be initialized with the length of the input
       * sequence.  The char pointer sequence_char points to a
       * dynamic char array of size length, and each char element has
       * already been set to ASCII value of the corresponding position
       * in the input amino acid sequence.  The function MapSequence will
       * convert the string of amino acid letter in the dynamic array the
       * char pointer sequence_char points to into a string of bits.  This
       * string of bits will be stored in the dynamic array the int
       * pointer sequence_bit points to.
       * ----------
       */

      sequence_int = MapSequence (sequence_char, length, map);

      if (name != NULL) {


	/*
	 * ----------
	 * Tom originally wrote:
	 *
	 * printf ("eMotif search results for: %s\n\n", name);
	 *
	 * The string "eMotif search results for:" is redundant and is
	 * eliminated.
	 * ----------
	 */

	printf (">%s\n", name);
      }  /* if (name != NULL) */


      /*
       * ----------
       * The following for loop will iterate through all the eMOTIFS in
       * the entire database.  For emotif-3.0, those
       * /usr/local/cnevill/emotif/emotifs/blocks.# files and
       * /usr/local/cnevill/emotif-2.0/data/blocks.# files no longer
       * exist.  Instead they have been merged into one big file called
       * emotif-3.0/emotifs/emotifs.txt.  The information contained in the
       * file emotif-3.0/emotifs/emotifs.txt is split into four .c files:
       * emotif-3.0/emotifs/emotifs.c, emotif-3.0/emotifs/emotifs.desc.c,
       * emotif-3.0/emotifs/emotifs.ind.c, and
       * emotif-3.0/emotifs/emotifs.spec.c.  The file
       * emotif-3.0/emotifs/emotifs.c contains the definition of the
       * static array the char* pointer motif points to.  The file
       * emotif-3.0/emotifs/emotifs.desc.c contains the definition of the
       * static array the char* pointer id points to.  The file
       * emotif-3.0/emotifs/emotifs.sens.c contains the definition of the
       * static array the char* pointer sensitivity points to.  The file
       * emotif-3.0/emotifs/emotifs.ind.c contains the definition of the
       * static array the int pointer motif_index points to and the
       * variable total.  The file emotifs.int.c contains the definition
       * of the static array the int pointer ints points to.  The file
       * emotifs.spec.c contains the definition of the static array the
       * double pointer specificity points to.  These six files are
       * automatically generated by the Perl script
       * emotif-3.0/build/emotif-to-c.pl based on the file
       * emotif-3.0/emotifs/emotifs.txt.  Each '\n'-delimited line in the
       * file emotif-3.0/emotifs/emotifs.txt corresponds to an eMOTIF
       * record.  An example of an eMOTIF record is shown below:
       *
       * -67.5814716871936\tPF00564B Octicosapeptide repeat proteins.\t
       * gcglapligfrykcvqcsnydlcqkcelahkhpehlmlrmptnngpgmvda\t12%\n
       *
       * Each field is delimited by a tab, '\t', character.  The first
       * field is the log specificity value of the eMOTIF; the second
       * field is the accession number of the protein family from which
       * the eMOTIF came and the name of that protein family; the fourth
       * field is the eMOTIF regular expression; and the fifth field is
       * the percentage of proteins in the parent alignment containing the
       * said eMOTIF regular expression.
       * 
       * The int variable total, defined in the first line of the file
       * emotif-3.0/emotifs/emotifs.ind.c, is the total number of eMOTIFS
       * in the current database.  The int variable motifn spans from 0 up
       * to, but not including, the value stored in the int variable total;
       * it serves as the index variable for the static array the int
       * pointer motif_index points to.
       * ----------
       */

      for (motifn = 0; 
	   motifn < total && specificity[motifn] < -cutoff_specificity;
	   motifn++) {


	/*
	 * ----------
	 * All of the integers in the static array the int pointer
	 * motif_index points to, except for the very last one, refer
	 * to elements that contain the corresponding eMOTIFS' first bit
	 * patterns in the static array the int pointer ints points to.
	 * The length of an eMOTIF is then simply the difference between
	 * the index number of the element containing the next
	 * (motifn + 1) eMOTIF's first bit pattern and the index number of
	 * the element containing the current (motifn) eMOTIF's first bit
	 * pattern in the static array the int pointer ints points to.
	 * For the very last eMOTIF in the database, its length is
	 * determined by substracting the index number of the element
	 * containing the final eMOTIF's first bit pattern from the total
	 * number of elements in the static array the int pointer ints
	 * points to; the size of the static array the int pointer ints
	 * points to is stored in the last element of the static array the
	 * int pointer motif_index points to.
	 * ----------
	 */

	width = motif_index[motifn + 1] - motif_index[motifn];


	/*
	 * ----------
	 * The following for loop iterates from position 0 of the input
	 * sequence up to, but not including, position length - width + 1
	 * of the input sequence.  This for loop stops at the
	 * (length - width)th position, because the last possible
	 * starting position in the input sequence that may be covered by
	 * a motif of length width is position length - width.
	 *
	 *
	 * Length:       |<-------       12       ------->|
	 *
	 * Sequence:     X  X  X  X  X  X  X  *  *  *  *  *
	 *
	 * Index:        0  1  2  3  4  5  6  7  8  9 10 11
	 *
	 * 
	 * In the illustration above the asterisks (*) are to indicate
	 * the amino acids that match the eMOTIF in the input sequence.
	 * If the input sequence is 12 amino acids long, and the motif is
	 * 5 amino acids long, the range of the int variable j, or
	 * the index variable for the input sequence, should be from 0 up
	 * to, but not including, length - width + 1, which, in the context
	 * of this example, is (12 - 5 + 1) = 8.  For the above example, the
	 * int variable j will be iterated from 0 through 7.
	 * ----------
	 */

	for (j = 0; j < length - width + 1; j++) {


	  /*
	   * ----------
	   * The following for loop performs a string comparsion between
	   * the current eMOTIF and a substring of the same length from
	   * the input sequence bit by bit.  The int pointer sequence_int
	   * points to a static array that contains the bit patterns of
	   * every amino acid letter in the input sequence.  The pointer
	   * ints points to a static array that contains the bit patterns
	   * of every amino acid letter from each eMOTIF in the entire
	   * database of eMOTIFS.  Because the bit patterns of one eMOTIF
	   * follow immediately the bit patterns of the previous eMOTIF
	   * in the static int array the pointer ints points to, the
	   * computer needs additional information to determine where one
	   * eMOTIF ends and the other one begins.  That additional
	   * information is stored in the static array the int pointer
	   * motif_index points to.  All of the integers in the static
	   * array the int pointer motif_index points to, except for the
	   * very last one, refer to elements that contain the
	   * corresponding eMOTIFS' first bit patterns in the static array
	   * the int pointer ints points to.  The length of an eMOTIF is
	   * then simply the difference between the index number of the
	   * element containing the next eMOTIF's first bit pattern and the
	   * index number of the element containing the current eMOTIF's
	   * first bit pattern in the static array the int pointer ints
	   * points to.  For the very last eMOTIF in the database, its 
	   * length is determined by substracting the index number of the
	   * element containing the final eMOTIF's first bit pattern from
	   * the total number of elements in the static array the int
	   * pointer ints points to; the size of the static array the int
	   * pointer ints points to is stored in the last element of the
	   * static array the int pointer motif_index points to.  The
	   * example below illustrates how all the various static int
	   * arrays and static char arrays are related:
	   *
	   * Each element in the static array the char* pointer motif points
	   * to is an eMOTIF string.  The string motif[0] points to is the
	   * the first eMOTIF, the string motif[1] points to is the second
	   * eMOTIF, and the string motif[54332] points to is the 54333rd
	   * eMOTIF.
	   *
	   * char *motif[] = {
	   *   "mnrheydcffiifgdcwpdayyhvfcakrtkttsefytagggltgvqnglaiagd",
	   *   "sllssgfspdfatvitmdrkaskqqtnidviyervrddyqlhwtstnwkgtntkd",
	   *   .
	   *   .
	   *   .
	   *   "[filvy]..........[fly].[filmv][ilmv]...y",
	   *   "m.l.[ilv]....[ilv]"};
	   * };
	   *
	   *
	   * Each int element in the static array the int pointer
	   * motif_index points to, except for the last one, is an index
	   * number for the static array the int pointer ints points to.
	   *
	   * int total = 54333;
	   *
	   * int motif_index[] = {
	   *   0,  --> ints[0] contains the first bit pattern for the
	   *           eMOTIF string the char pointer motif[0] points to.
	   *   55, --> ints[55] contains the first bit pattern for the
	   *           eMOTIF string the char pointer motif[1] points to.
	   *   110, --> ints[110] contains the first bit pattern for the
	   *            eMOTIF string the char pointer motif[2] points to.
	   *   .
	   *   .
	   *   .
	   *   917349, --> ints[917349] contains the first bit pattern for
	   *               the eMOTIF string the char pointer motif[54332]
	   *               points to.
	   *   917359 --> motif[54333] stores the size of the static int
	   *              array the pointer ints points to.
	   * };
	   *
	   * int ints[] = {
	   *   4096,  --> ints[0] contains the first bit pattern for the
	   *              55-amino-acid-long eMOTIF string the char
	   *              pointer motif[0] points to.
	   *   8192,  --> ints[1] contains the second bit pattern for the
	   *   131072,    55-amino-acid-long eMOTIF string the char
	   *              pointer motif[0] points to.
	   *   128,  --> ints[2] contains the third bit pattern for the
	   *             55-amino-acid-long eMOTIF string the char pointer
	   *             motif[0] points to.
	   *   .
	   *   .
	   *   .
	   * };
	   *
	   *
	   * The int variable k is an index variable for a given eMOTIF; it
	   * spans from 0 up to, but not including, the length of that
	   * eMOTIF.  The int variable j is an index variable for the input
	   * sequence; it spans from 0 up to, but not including, the length
	   * of that input sequence.  The int variable total, defined in
	   * the first line of the file emotif-3.0/emotifs/emotifs.ind.c,
	   * is the total number of eMOTIFS in the current database.  The
	   * int variable motifn spans from 0 up to, but not including,
	   * the value stored in the int variable total; it serves as the
	   * index variable for the static array the int pointer
	   * motif_index points to, the static array the char* pointer
	   * motif points to, and the static array the double pointer
	   * specificity points to.
	   *
	   * The bit-AND (&) operator compares two bits.  The result of this
	   * comparison is 1, if and only if both bits are 1.  When two
	   * variables are "bit-ANDed" together, as shown in the expression
	   * (sequence_int[j + k] & ints[motif_index[motifn] + k]), the
	   * bit-AND operator works on each bit independently.  The result
	   * of all the bit-AND operations is an integer in its own right,
	   * and this integer, in the for loop below, is used as a for-loop
	   * guard condition.  The only way for the for loop to proceed is
	   * if the integer that is the result of all the bit-AND operations
	   * is non-zero, provided that the index variable k has not reached
	   * the last character in the eMOTIF.
	   * ----------
	   */

	  for (k = 0;
	       k < width && 
		 (sequence_int[j + k] & ints[motif_index[motifn] + k]);
	       k++) ;


	  /* 
	   * ----------
	   * If the int variable k is equal to the width of current eMOTIF,
	   * stored in the int varialbe width, a region in the input
	   * sequence has been found to match the current eMOTIF in its
	   * entirety.  The width of an eMOTIF will have been calculated
	   * earlier in the program by the line of code:
	   *
	   * width = motif_index[motifn + 1] - motif_index[motifn];
	   *
	   * The private function CopyHit is called to extract the sequence
	   * segment that matches the current eMOTIF.
	   * ----------
	   */

	  if (k == width) {
	    CopyHit (segment, sequence_char, j, width, length);


	    /*
	     * ----------
	     * The rest of the emotif-search result is printed below.  The
	     * specificity of the eMOTIF is reported up to 3 significant
	     * figures, indicated by the number 3 preceding the format
	     * variable e and following the period.  The specificity
	     * reported here is in exponential format, whereas the only the
	     * first 12 digits from the double value the math.h-exported
	     * pow function will be displayed.  Because the eMOTIF's
	     * specificity was stored as a log value in the static array
	     * the double pointer specificity points to, the pow function
	     * was called with a base value of 10.0 to convert it to
	     * the exponential format.  The eMOTIF's sensitivity follows
	     * the eMOTIF's specificity.  The eMOTIF's accession
	     * number and the name of the protein family from which the
	     * eMOTIF came follows the eMOTIF's sensitivity.  Finally the
	     * actual eMOTIF regular expression is printed.
	     * ----------
	     */

	    printf ("%.10e|%s|%s|%s|", pow (10.0, specificity[motifn]),
		    sensitivity[motifn], id[motifn], motif[motifn]);


	    /*
	     * ----------
	     * Tom originally wrote:
	     *
	     * printf ("                        %d--%d\t%s\n\n", j + 1,
	     * j + width + 1, segment);
	     *
	     * If, for example, the input sequence contains a 10-amino-acid-
	     * long eMOTIF, and the region of the input sequence that
	     * matches the eMOTIF begins at residue 16 (j = 15; j + 1 = 16), 
	     * the last amino acid in that matching region would be residue
	     * 26 (j = 15; j + width + 1 = 15 + 10 + 1 = 26).  The last
	     * amino acid in that matching region should be residue 25
	     * (j = 15; j + width = 15 + 10 = 25).
	     * ----------
	     */

	    printf("%s|%d|%d\n", segment, j + 1, j + width);
	  }  /* if (k == width) */
	}  /* for (j = 0; j < length - width + 1; j++) */
      }  /* for (motifn = 0; 
	    motifn < total && specificity[motifn] < -cutoff_specificity;
	    motifn++) */

    
      /*
       * ----------
       * The function MapSequence allocated a block memory dynamically
       * earlier in the program.  The int pointer sequence_int points to
       * that block of dynamically allocated memory.  To reclaim that block
       * of memory, it must be freed explicitly using the free function
       * exported by the stdlib.h file.
       * ----------
       */

      free (sequence_int);


      /*
       * ----------
       * The function Sequence_read, exported by the file sequence.h,
       * has allocated a block memory dynamically earlier in the program.
       * The char pointer sequence_char points to that block of dynamically
       * allocated memory.  To reclaim that block of memory, it must be
       * freed explicitly using the free function exported by the stdlib.h
       * file.
       * ----------
       */

      free (sequence_char);
    }  /* if (sequence_char = Sequence_read (fp, &name, &length)) */
  }  /* while (!feof (fp) && !isspace(c = fgetc (fp))) */


  /*
   * ----------
   * For the Sun SPARC machines the function main must return 0 to indicate
   * that the program has reached completion; this return 0 is optional
   * for SGI machines.
   * ----------
   */

  return (0);
}
